"""Client."""
from __future__ import annotations

import asyncio
from collections import defaultdict
from copy import deepcopy
from datetime import datetime
from functools import partial
import json
import logging
from operator import itemgetter
import pprint
from types import TracebackType
from typing import Any, DefaultDict, Dict, List
import uuid

from aiohttp import ClientSession, ClientWebSocketResponse, WSMsgType, client_exceptions

from ..common.json_utils import CHIPJSONDecoder, CHIPJSONEncoder
from ..common.model.message import (
    CommandMessage,
    ErrorResultMessage,
    Message,
    ResultMessage,
    ServerInformation,
    SubscriptionReportMessage,
    SuccessResultMessage,
)
from ..common.model.version import VersionInfo
from .const import MAX_SERVER_SCHEMA_VERSION, MIN_SERVER_SCHEMA_VERSION
from .exceptions import (
    CannotConnect,
    ConnectionClosed,
    ConnectionFailed,
    FailedCommand,
    InvalidMessage,
    InvalidServerVersion,
    InvalidState,
    NotConnected,
)
from .model.driver import Driver

SIZE_PARSE_JSON_EXECUTOR = 8192

# Message IDs
SET_API_SCHEMA_MESSAGE_ID = "api-schema-id"
GET_INITIAL_LOG_CONFIG_MESSAGE_ID = "get-initial-log-config"
START_LISTENING_MESSAGE_ID = "listen-id"

LISTEN_MESSAGE_IDS = (
    GET_INITIAL_LOG_CONFIG_MESSAGE_ID,
    SET_API_SCHEMA_MESSAGE_ID,
    START_LISTENING_MESSAGE_ID,
)


class Client:
    """Class to manage the connection to the Matter server."""

    def __init__(
        self,
        ws_server_url: str,
        aiohttp_session: ClientSession,
        schema_version: int = MAX_SERVER_SCHEMA_VERSION,
        record_messages: bool = False,
    ):
        """Initialize the Client class."""
        self.ws_server_url = ws_server_url
        self.aiohttp_session = aiohttp_session
        self.driver: Driver | None = None
        # The WebSocket client
        self._client: ClientWebSocketResponse | None = None
        # Version of the connected server
        self.version: VersionInfo | None = None
        self.server_info: ServerInformation | None = None
        self.schema_version: int = schema_version
        self.logger = logging.getLogger(__package__)
        self._loop = asyncio.get_running_loop()
        self._result_futures: Dict[str, asyncio.Future] = {}
        self._shutdown_complete_event: asyncio.Event | None = None
        self._record_messages = record_messages
        self._recorded_commands: DefaultDict[str, dict] = defaultdict(dict)
        self._recorded_events: List[dict] = []

    def __repr__(self) -> str:
        """Return the representation."""
        prefix = "" if self.connected else "not "
        return f"{type(self).__name__}(ws_server_url={self.ws_server_url!r}, {prefix}connected)"

    @property
    def connected(self) -> bool:
        """Return if we're currently connected."""
        return self._client is not None and not self._client.closed

    @property
    def recording_messages(self) -> bool:
        """Return True if messages are being recorded."""
        return self._record_messages

    async def async_send_command(
        self,
        command: str,
        args: dict[str, Any],
        require_schema: int | None = None,
    ) -> dict:
        """Send a command and get a response."""
        if require_schema is not None and require_schema > self.schema_version:
            raise InvalidServerVersion(
                "Command not available due to incompatible server version. Update the Z-Wave "
                f"JS Server to a version that supports at least api schema {require_schema}."
            )

        message = CommandMessage(
            messageId=uuid.uuid4().hex,
            command=command,
            args=args,
        )
        future: asyncio.Future[dict] = self._loop.create_future()
        self._result_futures[message.messageId] = future
        await self._send_message(message)
        try:
            return await future
        finally:
            self._result_futures.pop(message.messageId)

    async def async_send_command_no_wait(
        self, command: str, args: dict[str, Any], require_schema: int | None = None
    ) -> None:
        """Send a command without waiting for the response."""
        if require_schema is not None and require_schema > self.schema_version:
            raise InvalidServerVersion(
                "Command not available due to incompatible server version. Update the Matter "
                f"Server to a version that supports at least api schema {require_schema}."
            )
        message = CommandMessage(
            messageId=uuid.uuid4().hex,
            command=command,
            args=args,
        )
        await self._send_message(message)

    async def connect(self) -> None:
        """Connect to the websocket server."""
        if self.driver is not None:
            raise InvalidState("Re-connected with existing driver")

        self.logger.debug("Trying to connect")
        try:
            self._client = await self.aiohttp_session.ws_connect(
                self.ws_server_url,
                heartbeat=55,
                compress=15,
                max_msg_size=0,
            )
        except (
            client_exceptions.WSServerHandshakeError,
            client_exceptions.ClientError,
        ) as err:
            raise CannotConnect(err) from err

        self.version = version = await self._receive_message_or_raise()

        # basic check for server schema version compatibility
        if (
            self.version.min_schema_version > MIN_SERVER_SCHEMA_VERSION
            or self.version.max_schema_version < MIN_SERVER_SCHEMA_VERSION
        ):
            await self._client.close()
            raise InvalidServerVersion(
                f"Matter Server version is incompatible: {self.version.server_version} "
                "a version is required that supports at least "
                f"api schema {MIN_SERVER_SCHEMA_VERSION}"
            )
        # store the (highest possible) schema version we're going to use/request
        # this is a bit future proof as we might decide to use a pinned version at some point
        # for now we just negotiate the highest available schema version and
        # guard incompatibility with the MIN_SERVER_SCHEMA_VERSION
        if self.version.max_schema_version < MAX_SERVER_SCHEMA_VERSION:
            self.schema_version = self.version.max_schema_version

        await self._send_message(
            CommandMessage(
                messageId="server-get-info",
                command="server.GetInfo",
                args={},
            )
        )
        msg = await self._receive_message_or_raise()

        if (
            not isinstance(msg, SuccessResultMessage)
            or not isinstance(msg.result, ServerInformation)
            or msg.messageId != "server-get-info"
        ):
            raise InvalidMessage(
                "Expected a SuccessResultMessage with messageId 'server-get-info'"
            )

        self.server_info = msg.result

        self.logger.info(
            "Connected to Server %s, Driver %s, Using Schema %s",
            version.server_version,
            version.driver_version,
            self.schema_version,
        )

    async def listen(self, driver_ready: asyncio.Event) -> None:
        """Start listening to the websocket."""
        if not self.connected:
            raise InvalidState("Not connected when start listening")

        assert self._client

        try:
            self.driver = Driver(self)

            self.logger.info("Matter initialized.")
            driver_ready.set()

            while not self._client.closed:
                data = await self._receive_message_or_raise()

                self._handle_incoming_message(data)
        except ConnectionClosed:
            pass

        finally:
            self.logger.debug("Listen completed. Cleaning up")

            for future in self._result_futures.values():
                future.cancel()

            if not self._client.closed:
                await self._client.close()

            if self._shutdown_complete_event:
                self._shutdown_complete_event.set()

    async def disconnect(self) -> None:
        """Disconnect the client."""
        self.logger.debug("Closing client connection")

        if not self.connected:
            return

        assert self._client

        # 'listen' was never called
        if self.driver is None:
            await self._client.close()
            return

        self._shutdown_complete_event = asyncio.Event()
        await self._client.close()
        await self._shutdown_complete_event.wait()

        self._shutdown_complete_event = None
        self.driver = None

    def begin_recording_messages(self) -> None:
        """Begin recording messages for replay later."""
        if self._record_messages:
            raise InvalidState("Already recording messages")

        self._record_messages = True

    def end_recording_messages(self) -> List[dict]:
        """End recording messages and return messages that were recorded."""
        if not self._record_messages:
            raise InvalidState("Not recording messages")

        self._record_messages = False

        data = sorted(
            (*self._recorded_commands.values(), *self._recorded_events),
            key=itemgetter("ts"),
        )
        self._recorded_commands.clear()
        self._recorded_events.clear()

        return list(data)

    async def _receive_message_or_raise(self) -> Message:
        """Receive message or raise."""
        assert self._client
        msg = await self._client.receive()

        if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
            raise ConnectionClosed("Connection was closed.")

        if msg.type == WSMsgType.ERROR:
            raise ConnectionFailed()

        if msg.type != WSMsgType.TEXT:
            raise InvalidMessage(f"Received non-Text message: {msg.type}")

        try:
            if len(msg.data) > SIZE_PARSE_JSON_EXECUTOR:
                obj: dict = await self._loop.run_in_executor(
                    None, lambda: json.loads(msg.data, cls=CHIPJSONDecoder)
                )
            else:
                obj = json.loads(msg.data, cls=CHIPJSONDecoder)
        except ValueError as err:
            raise InvalidMessage("Received invalid JSON.") from err

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug("Received message:\n%s\n", pprint.pformat(msg))

        return obj

    def _handle_incoming_message(self, msg: Any) -> None:
        """Handle incoming message.

        Run all async tasks in a wrapper to log appropriately.
        """
        if isinstance(msg, ResultMessage):
            msg: ResultMessage = msg
            future = self._result_futures.get(msg.messageId)

            if future is None:
                # no listener for this result
                return

            if self._record_messages and msg.messageId not in LISTEN_MESSAGE_IDS:
                self._recorded_commands[msg.messageId].update(
                    {
                        "result_ts": datetime.utcnow().isoformat(),
                        "result_msg": deepcopy(msg),
                    }
                )

            if isinstance(msg, SuccessResultMessage):
                msg: SuccessResultMessage = msg
                future.set_result(msg.result)
                return
            elif isinstance(msg, ErrorResultMessage):
                msg: ErrorResultMessage = msg
                future.set_exception(FailedCommand(msg.messageId, msg.errorCode))
            else:
                raise InvalidMessage("Invalid ResultMessage.")

            return
        elif isinstance(msg, SubscriptionReportMessage):
            self.logger.debug(
                "Received subscription report: %s",
                msg,
            )

            if self._record_messages:
                self._recorded_events.append(
                    {
                        "record_type": "subscription_report",
                        "ts": datetime.utcnow().isoformat(),
                        "payload": deepcopy(msg),
                    }
                )

            self.driver.read_subscriptions.receive_event(msg.payload)
        else:
            # Can't handle
            self.logger.debug(
                "Received message with unknown type '%s': %s",
                type(msg),
                msg,
            )
            return

    async def _send_message(self, message: CommandMessage) -> None:
        """Send a message.

        Raises NotConnected if client not connected.
        """
        if not self.connected:
            raise NotConnected

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug("Publishing message:\n%s\n", pprint.pformat(message))

        assert self._client
        assert isinstance(message, CommandMessage)

        if self._record_messages and message.messageId not in LISTEN_MESSAGE_IDS:
            # We don't need to deepcopy command_msg because it is always released by
            # the caller after the command is sent.
            self._recorded_commands[message.messageId].update(
                {
                    "record_type": "command",
                    "ts": datetime.utcnow().isoformat(),
                    "command": message["command"],
                    "command_msg": message,
                }
            )

        await self._client.send_json(
            message, dumps=partial(json.dumps, cls=CHIPJSONEncoder)
        )

    async def __aenter__(self) -> "Client":
        """Connect to the websocket."""
        await self.connect()
        return self

    async def __aexit__(
        self, exc_type: Exception, exc_value: str, traceback: TracebackType
    ) -> None:
        """Disconnect from the websocket."""
        await self.disconnect()



#
#    Copyright (c) 2020 Project CHIP Authors
#    Copyright (c) 2020 Google LLC.
#    All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

#
#    @file
#      This file is utility for Chip
#

from __future__ import absolute_import
from __future__ import print_function
import binascii
from ctypes import *


class ChipUtility(object):
    @staticmethod
    def Hexlify(val):
        return binascii.hexlify(val).decode()

    @staticmethod
    def VoidPtrToByteArray(ptr, len):
        if ptr:
            v = bytearray(len)
            memmove((c_byte * len).from_buffer(v), ptr, len)
            return v
        return None

    @staticmethod
    def ByteArrayToVoidPtr(array):
        if array != None:
            if not (isinstance(array, bytes) or isinstance(array, bytearray)):
                raise TypeError("Array must be an str or a bytearray")
            return cast((c_byte * len(array)).from_buffer_copy(array), c_void_p)
        return c_void_p(0)

    @staticmethod
    def IsByteArrayAllZeros(array):
        for i in array:
            if i != 0:
                return False
        return True

    @staticmethod
    def ByteArrayToHex(array):
        return ChipUtility.Hexlify(bytes(array))

    @staticmethod
    def CStringToString(s):
        return None if s is None else s.decode()

    @staticmethod
    def StringToCString(s):
        return None if s is None else s.encode()

# To support Python 3.8 and older versions, which does not support @classmethod + @property


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()



#
#    Copyright (c) 2020 Project CHIP Authors
#    All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from dataclasses import dataclass, asdict, field, make_dataclass
from typing import ClassVar, List, Dict, Any, Mapping, Type, Union, ClassVar
import enum
import typing
from matter_server.vendor.chip import tlv, ChipUtility
from matter_server.vendor.chip.clusters.Types import Nullable, NullValue
from dacite import from_dict


def GetUnionUnderlyingType(typeToCheck, matchingType=None):
    """This retrieves the underlying types behind a unioned type by appropriately
    passing in the required matching type in the matchingType input argument.

    If that is 'None' (not to be confused with NoneType), then it will retrieve
    the 'real' type behind the union, i.e not Nullable && not None
    """
    if not (typing.get_origin(typeToCheck) == typing.Union):
        return None

    for t in typing.get_args(typeToCheck):
        if matchingType is None:
            if t != type(None) and t != Nullable:
                return t
        else:
            if t == matchingType:
                return t

    return None


@dataclass
class ClusterObjectFieldDescriptor:
    Label: str = ""
    Tag: int = None
    Type: Type = None

    def _PutSingleElementToTLV(
        self, tag, val, elementType, writer: tlv.TLVWriter, debugPath: str = "?"
    ):
        if issubclass(elementType, ClusterObject):
            if not isinstance(val, dict):
                raise ValueError(
                    f"Field {debugPath}.{self.Label} expected a struct, but got {type(val)}"
                )
            elementType.descriptor.DictToTLVWithWriter(
                f"{debugPath}.{self.Label}", tag, val, writer
            )
            return

        try:
            val = elementType(val)
        except Exception:
            raise ValueError(
                f"Field {debugPath}.{self.Label} expected {elementType}, but got {type(val)}"
            )
        writer.put(tag, val)

    def PutFieldToTLV(self, tag, val, writer: tlv.TLVWriter, debugPath: str = "?"):
        if val == NullValue:
            if GetUnionUnderlyingType(self.Type, Nullable) is None:
                raise ValueError(
                    f"Field {debugPath}.{self.Label} was not nullable, but got a null"
                )

            writer.put(tag, None)
        elif val is None:
            if GetUnionUnderlyingType(self.Type, type(None)) is None:
                raise ValueError(
                    f"Field {debugPath}.{self.Label} was not optional, but encountered None"
                )
        else:
            #
            # If it is an optional or nullable type, it's going to be a union.
            # So, let's get at the 'real' type within that union before proceeding,
            # since at this point, we're guarenteed to not get None or Null as values.
            #
            elementType = GetUnionUnderlyingType(self.Type)
            if elementType is None:
                elementType = self.Type

            if not isinstance(val, List):
                self._PutSingleElementToTLV(tag, val, elementType, writer, debugPath)
                return

            writer.startArray(tag)

            # Get the type of the list. This is a generic, which has its sub-type information of the list element
            # inside its type argument.
            (elementType,) = typing.get_args(elementType)

            for i, v in enumerate(val):
                self._PutSingleElementToTLV(
                    None, v, elementType, writer, debugPath + f"[{i}]"
                )
            writer.endContainer()


@dataclass
class ClusterObjectDescriptor:
    Fields: List[ClusterObjectFieldDescriptor]

    def GetFieldByTag(self, tag: int) -> ClusterObjectFieldDescriptor:
        for field in self.Fields:
            if field.Tag == tag:
                return field
        return None

    def GetFieldByLabel(self, label: str) -> ClusterObjectFieldDescriptor:
        for field in self.Fields:
            if field.Label == label:
                return field
        return None

    def _ConvertNonArray(self, debugPath: str, elementType, value: Any) -> Any:
        if not issubclass(elementType, ClusterObject):
            if issubclass(elementType, enum.Enum):
                value = elementType(value)

            if not isinstance(value, elementType):
                raise ValueError(
                    f"Failed to decode field {debugPath}, expected type {elementType}, got {type(value)}"
                )
            return value
        if not isinstance(value, Mapping):
            raise ValueError(f"Failed to decode field {debugPath}, struct expected.")
        return elementType.descriptor.TagDictToLabelDict(debugPath, value)

    def TagDictToLabelDict(
        self, debugPath: str, tlvData: Dict[int, Any]
    ) -> Dict[str, Any]:
        ret = {}
        for tag, value in tlvData.items():
            descriptor = self.GetFieldByTag(tag)
            if not descriptor:
                # We do not have enough information for this field.
                ret[tag] = value
                continue

            if value is None:
                ret[descriptor.Label] = NullValue
                continue

            if typing.get_origin(descriptor.Type) == typing.Union:
                realType = GetUnionUnderlyingType(descriptor.Type)
                if realType is None:
                    raise ValueError(
                        f"Field {debugPath}.{self.Label} has no valid underlying data model type"
                    )

                valueType = realType
            else:
                valueType = descriptor.Type

            if typing.get_origin(valueType) == list:
                listElementType = typing.get_args(valueType)[0]
                ret[descriptor.Label] = [
                    self._ConvertNonArray(f"{debugPath}[{i}]", listElementType, v)
                    for i, v in enumerate(value)
                ]
                continue
            ret[descriptor.Label] = self._ConvertNonArray(
                f"{debugPath}.{descriptor.Label}", valueType, value
            )
        return ret

    def TLVToDict(self, tlvBuf: bytes) -> Dict[str, Any]:
        tlvData = tlv.TLVReader(tlvBuf).get().get("Any", {})
        return self.TagDictToLabelDict([], tlvData)

    def DictToTLVWithWriter(
        self, debugPath: str, tag, data: Mapping, writer: tlv.TLVWriter
    ):
        writer.startStructure(tag)
        for field in self.Fields:
            val = data.get(field.Label, None)
            field.PutFieldToTLV(field.Tag, val, writer, debugPath + f".{field.Label}")
        writer.endContainer()

    def DictToTLV(self, data: dict) -> bytes:
        tlvwriter = tlv.TLVWriter(bytearray())
        self.DictToTLVWithWriter("", None, data, tlvwriter)
        return bytes(tlvwriter.encoding)


class ClusterObject:
    def ToTLV(self):
        return self.descriptor.DictToTLV(asdict(self))

    @classmethod
    def FromDict(cls, data: dict):
        return from_dict(data_class=cls, data=data)

    @classmethod
    def FromTLV(cls, data: bytes):
        return cls.FromDict(data=cls.descriptor.TLVToDict(data))

    @ChipUtility.classproperty
    def descriptor(cls):
        raise NotImplementedError()


class ClusterCommand(ClusterObject):
    @ChipUtility.classproperty
    def cluster_id(self) -> int:
        raise NotImplementedError()

    @ChipUtility.classproperty
    def command_id(self) -> int:
        raise NotImplementedError()

    @ChipUtility.classproperty
    def must_use_timed_invoke(cls) -> bool:
        return False


class Cluster(ClusterObject):
    """
    When send read requests with returnClusterObject=True, we will set the data_version property of the object.
    Otherwise the [endpoint][cluster][Clusters.DataVersion] will be set to the DataVersion of the cluster.

    For data_version, we do not make it a real property so we can distinguish it with real attributes internally,
    especially the TLV decoding logic. Also ThreadNetworkDiagnostics has an attribute with the same name so we
    picked data_version as its name.
    """

    @property
    def data_version(self) -> int:
        return self._data_version

    def __rich_repr__(self):
        """
        Override the default behavior of rich.pretty.pprint for adding the cluster data version.
        """
        if self._data_version is not None:
            yield "(data version)", self.data_version
        for k in self.__dataclass_fields__.keys():
            if k in self.__dict__:
                yield k, self.__dict__[k]

    def SetDataVersion(self, version: int) -> None:
        self._data_version = version


class ClusterAttributeDescriptor:
    """
    The ClusterAttributeDescriptor is used for holding an attribute's metadata like its cluster id, attribute id and its type.

    Users should not initialize an object based on this class. Instead, users should pass the subclass objects to tell some methods what they want.

    The implementation of this functions is quite tricky, it will create a cluster object on-the-fly, and use it for actual encode / decode routine to save lines of code.
    """

    @classmethod
    def ToTLV(cls, tag: Union[int, None], value):
        writer = tlv.TLVWriter()
        wrapped_value = cls._cluster_object(Value=value)
        cls.attribute_type.PutFieldToTLV(
            tag, asdict(wrapped_value)["Value"], writer, ""
        )
        return writer.encoding

    @classmethod
    def FromTLV(cls, tlvBuffer: bytes):
        obj_class = cls._cluster_object
        return obj_class.FromDict(
            obj_class.descriptor.TagDictToLabelDict(
                "", {0: tlv.TLVReader(tlvBuffer).get().get("Any", {})}
            )
        ).Value

    @classmethod
    def FromTagDictOrRawValue(cls, val: Any):
        obj_class = cls._cluster_object
        return obj_class.FromDict(
            obj_class.descriptor.TagDictToLabelDict("", {0: val})
        ).Value

    @ChipUtility.classproperty
    def cluster_id(self) -> int:
        raise NotImplementedError()

    @ChipUtility.classproperty
    def attribute_id(self) -> int:
        raise NotImplementedError()

    @ChipUtility.classproperty
    def attribute_type(cls) -> ClusterObjectFieldDescriptor:
        raise NotImplementedError()

    @ChipUtility.classproperty
    def must_use_timed_write(cls) -> bool:
        return False

    @ChipUtility.classproperty
    def _cluster_object(cls) -> ClusterObject:
        return make_dataclass(
            "InternalClass",
            [
                ("Value", cls.attribute_type.Type, field(default=None)),
                (
                    "descriptor",
                    ClassVar[ClusterObjectDescriptor],
                    field(
                        default=ClusterObjectDescriptor(
                            Fields=[
                                ClusterObjectFieldDescriptor(
                                    Label="Value", Tag=0, Type=cls.attribute_type.Type
                                )
                            ]
                        )
                    ),
                ),
            ],
            bases=(ClusterObject,),
        )


class ClusterEvent(ClusterObject):
    @ChipUtility.classproperty
    def cluster_id(self) -> int:
        raise NotImplementedError()

    @ChipUtility.classproperty
    def event_id(self) -> int:
        raise NotImplementedError()



import argparse
import asyncio
import logging
import os
from pathlib import Path
import sys

import coloredlogs

from .matter_stack import MatterStack
from .server import MatterServer

_LOGGER = logging.getLogger(__package__)


def get_arguments() -> argparse.Namespace:
    """Get parsed passed in arguments."""

    parser = argparse.ArgumentParser(
        description="Matter Controller Server using WebSockets."
    )
    parser.add_argument(
        "--log-file",
        type=str,
        default=None,
        help="Log file to write to.  If not set, matter_server.log is used",
    )
    parser.add_argument(
        "-log",
        "--log-level",
        default="info",
        help="Provide logging level. Example --log-level debug, default=info, possible=(critical, error, warning, info, debug)",
    )

    arguments = parser.parse_args()

    return arguments


def main() -> int:
    args = get_arguments()
    handlers = None

    if args.log_file:
        handlers = [logging.FileHandler(args.log_file)]

    logging.basicConfig(handlers=handlers, level=args.log_level.upper())

    host = os.getenv("CHIP_WS_SERVER_HOST", "::,0.0.0.0").split(",")
    port = int(os.getenv("CHIP_WS_SERVER_PORT", "5580"))
    storage_path = os.getenv(
        "CHIP_WS_STORAGE",
        str(Path.home() / ".chip-storage/python-kv.json"),
    )
    debug = os.getenv("CHIP_WS_DEBUG") is not None

    coloredlogs.install(level=logging.DEBUG if debug else args.log_level.upper())
    stack = MatterStack(storage_path)

    loop = asyncio.get_event_loop()

    async def create_server():
        return MatterServer(stack)

    server = loop.run_until_complete(create_server())
    server.run(host, port)
    stack.shutdown()


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, Callable

from matter_server.client.exceptions import NotConnected

if TYPE_CHECKING:
    from ..client import Client
    from .device_controller import (
        ReadAttributesType,
        ReadDataVersionFiltersType,
        ReadEventsType,
    )


class ReadSubscriptions:
    def __init__(self, client: Client):
        self.client = client
        self.subscriptions: dict[int, Callable[[dict], None]] = {}

    async def subscribe_node(
        self,
        nodeid: int,
        subscription_callback: Callable[[dict], None],
        reportInterval: tuple[int, int],
        attributes: ReadAttributesType = None,
        dataVersionFilters: ReadDataVersionFiltersType = None,
        events: ReadEventsType = None,
        returnClusterObject: bool = False,
        fabricFiltered: bool = True,
        keepSubscriptions: bool = False,
    ):
        """Subscribe to a node."""
        read_result = await self.client.driver.device_controller.read(
            nodeid=nodeid,
            reportInterval=reportInterval,
            attributes=attributes,
            dataVersionFilters=dataVersionFilters,
            events=events,
            returnClusterObject=returnClusterObject,
            fabricFiltered=fabricFiltered,
            keepSubscriptions=keepSubscriptions,
        )

        self.subscriptions[read_result["subscription_id"]] = subscription_callback

        async def unsubscribe():
            with suppress(NotConnected):
                await self.client.async_send_command(
                    "device_controller.Unsubscribe",
                    {"subscription_id": read_result["subscription_id"]},
                )

        return unsubscribe

    def receive_event(self, event):
        """Receive a subscription event."""
        subscription_id = event["subscriptionId"]
        if not (callback := self.subscriptions.get(subscription_id)):
            self.client.logger.warning(
                f"No Subscription object for Subscription Id 0x{subscription_id:X} present."
            )
            return
        try:
            callback(event)
        except Exception:  # pylint: disable=broad-except
            self.client.logger.exception(
                "Unexpected exception handling event %s", event
            )



"""Matter node."""
from __future__ import annotations

from typing import TYPE_CHECKING

from matter_server.vendor import device_types
from matter_server.vendor.chip.clusters import Objects as all_clusters

from .device_type_instance import MatterDeviceTypeInstance
from .node_device import (
    AbstractMatterNodeDevice,
    MatterBridgedNodeDevice,
    MatterNodeDevice,
)

if TYPE_CHECKING:
    from ..matter import Matter


class MatterNode:
    """Matter node."""

    root_device_type_instance: MatterDeviceTypeInstance[device_types.RootNode]
    bridge_device_type_instance: MatterDeviceTypeInstance[
        device_types.Bridge
    ] | None = None
    device_type_instances: list[MatterDeviceTypeInstance]
    node_devices: list[AbstractMatterNodeDevice]

    def __init__(self, matter: Matter, node_info: dict) -> None:
        self.matter = matter
        self.raw_data = node_info

        device_type_instances: list[MatterDeviceTypeInstance] = []

        for endpoint_id, endpoint_info in node_info["attributes"].items():
            descriptor: all_clusters.Descriptor = endpoint_info["Descriptor"]
            for device_info in descriptor.deviceList:
                device_type = device_types.ALL_TYPES.get(device_info.type)

                if device_type is None:
                    matter.adapter.logger.warning(
                        "Found unknown device type %s", device_info.type
                    )
                    continue

                instance = MatterDeviceTypeInstance(
                    self, device_type, int(endpoint_id), device_info.revision
                )
                if device_type is device_types.RootNode:
                    self.root_device_type_instance = instance
                elif device_type is device_types.Bridge:
                    self.bridge_device_type_instance = instance
                else:
                    device_type_instances.append(instance)

        self.device_type_instances = device_type_instances

        if not hasattr(self, "root_device_type_instance"):
            raise ValueError("No root device found")

        self.node_devices = []

        if self.bridge_device_type_instance:
            for instance in device_type_instances:
                if instance.device_type == device_types.BridgedDevice:
                    self.node_devices.append(MatterBridgedNodeDevice(instance))

        else:
            self.node_devices.append(MatterNodeDevice(self))

    @property
    def node_id(self) -> int:
        return self.raw_data["node_id"]

    @property
    def name(self) -> str:
        return self.root_device_type_instance.get_cluster(all_clusters.Basic).nodeLabel

    @property
    def unique_id(self) -> str:
        return self.root_device_type_instance.get_cluster(all_clusters.Basic).uniqueID

    def update_data(self, node_info):
        self.raw_data = node_info

    def __repr__(self):
        return f"<MatterNode {self.node_id}>"



from __future__ import annotations

import asyncio
from concurrent import futures
from contextlib import suppress
import json
import logging
from typing import TYPE_CHECKING, Any, Callable, Final
import weakref

from aiohttp import WSMsgType, web
import async_timeout

from ..common.json_utils import CHIPJSONDecoder, CHIPJSONEncoder
from ..common.model.message import CommandMessage
from .active_connection import ActiveConnection, Disconnect

if TYPE_CHECKING:
    from .server import MatterServer

MAX_PENDING_MSG = 512
CANCELLATION_ERRORS: Final = (asyncio.CancelledError, futures.CancelledError)


def mount_websocket(server: MatterServer, path: str) -> None:
    """Mount the websocket endpoint."""
    clients = weakref.WeakSet()

    async def _handle_chip_ws(request):
        connection = WebSocketHandler(server, request)
        try:
            clients.add(connection)
            return await connection.async_handle()
        finally:
            clients.remove(connection)

    async def _handle_shutdown(app):
        for client in set(clients):
            await client.disconnect()

    server.app.on_shutdown.append(_handle_shutdown)
    server.app.router.add_route("GET", path, _handle_chip_ws)


class WebSocketAdapter(logging.LoggerAdapter):
    """Add connection id to websocket messages."""

    def process(self, msg: str, kwargs: Any) -> tuple[str, Any]:
        """Add connid to websocket log messages."""
        return f'[{self.extra["connid"]}] {msg}', kwargs


class WebSocketHandler:
    """Handle an active websocket client connection."""

    def __init__(self, server: MatterServer, request: web.Request) -> None:
        """Initialize an active connection."""
        self.server = server
        self.request = request
        self.wsock = web.WebSocketResponse(heartbeat=55)
        self._to_write: asyncio.Queue = asyncio.Queue(maxsize=MAX_PENDING_MSG)
        self._handle_task: asyncio.Task | None = None
        self._writer_task: asyncio.Task | None = None
        self._logger = WebSocketAdapter(server.logger, {"connid": id(self)})

    async def _writer(self) -> None:
        """Write outgoing messages."""
        # Exceptions if Socket disconnected or cancelled by connection handler
        with suppress(RuntimeError, ConnectionResetError, *CANCELLATION_ERRORS):
            while not self.wsock.closed:
                if (process := await self._to_write.get()) is None:
                    break

                if not isinstance(process, str):
                    message: str = process()
                else:
                    message = process
                self._logger.debug("Sending %s", message)
                await self.wsock.send_str(message)

    def _send_message(self, message: str | dict[str, Any] | Callable[[], str]) -> None:
        """Send a message to the client.

        Closes connection if the client is not reading the messages.

        Async friendly.
        """
        message = json.dumps(message, cls=CHIPJSONEncoder)

        try:
            self._to_write.put_nowait(message)
        except asyncio.QueueFull:
            self._logger.error(
                "Client exceeded max pending messages: %s", MAX_PENDING_MSG
            )

            self._cancel()

    def _cancel(self) -> None:
        """Cancel the connection."""
        if self._handle_task is not None:
            self._handle_task.cancel()
        if self._writer_task is not None:
            self._writer_task.cancel()

    async def disconnect(self):
        self._cancel()
        await self._writer_task

    async def async_handle(self) -> web.WebSocketResponse:
        """Handle a websocket response."""
        request = self.request
        wsock = self.wsock
        try:
            async with async_timeout.timeout(10):
                await wsock.prepare(request)
        except asyncio.TimeoutError:
            self._logger.warning("Timeout preparing request from %s", request.remote)
            return wsock

        self._logger.debug("Connected from %s", request.remote)
        self._handle_task = asyncio.current_task()

        self._writer_task = asyncio.create_task(self._writer())

        connection = ActiveConnection(self._logger, self.server, self._send_message)
        await connection.async_initialize()
        disconnect_warn = None

        try:
            while not wsock.closed:
                msg = await wsock.receive()

                if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSING):
                    break

                if msg.type != WSMsgType.TEXT:
                    disconnect_warn = "Received non-Text message."
                    break

                self._logger.debug("Received: %s", msg.data)

                try:
                    command_msg = json.loads(msg.data, cls=CHIPJSONDecoder)
                except ValueError:
                    disconnect_warn = "Received invalid JSON."
                    break

                self._logger.info("Deserialized message: %s", command_msg)

                if not isinstance(command_msg, CommandMessage):
                    disconnect_warn = "Did not receive a CommandMessage."
                    break

                self._logger.debug("Received %s", command_msg)
                connection.async_handle(command_msg)

        except asyncio.CancelledError:
            self._logger.info("Connection closed by client")

        except Disconnect:
            pass

        except Exception:  # pylint: disable=broad-except
            self._logger.exception("Unexpected error inside websocket API")

        finally:
            connection.async_handle_close()

            try:
                self._to_write.put_nowait(None)
                # Make sure all error messages are written before closing
                await self._writer_task
                await wsock.close()
            except asyncio.QueueFull:  # can be raised by put_nowait
                self._writer_task.cancel()

            finally:
                if disconnect_warn is None:
                    self._logger.debug("Disconnected")
                else:
                    self._logger.warning("Disconnected: %s", disconnect_warn)

        return wsock



import asyncio
import logging
import os
import sys

import aiohttp

from matter_server.client.client import Client
from matter_server.vendor.chip.clusters import Objects as clusters

logging.basicConfig(level=logging.DEBUG)
_LOGGER = logging.getLogger(__name__)

DEFAULT_HOST = "127.0.0.1" if len(sys.argv) < 2 else sys.argv[1]
DEFAULT_PORT = "5580" if len(sys.argv) < 3 else sys.argv[2]
HOST = os.getenv("CHIP_WS_SERVER_HOST", DEFAULT_HOST)
PORT = int(os.getenv("CHIP_WS_SERVER_PORT", DEFAULT_PORT))
URL = f"http://{HOST}:{PORT}/chip_ws"


async def main():
    async with aiohttp.ClientSession() as session:
        await connect(session)


async def connect(session):
    client = Client(URL, session)
    await client.connect()
    is_initialized = asyncio.Event()

    toggle_task = asyncio.create_task(toggle_happiness(client, is_initialized))

    try:
        await client.listen(is_initialized)
    finally:
        toggle_task.cancel()
        await toggle_task


async def toggle_happiness(client: Client, is_initialized: asyncio.Event):
    try:
        await is_initialized.wait()

        nodeid = 4335

        await client.driver.device_controller.resolve_node(nodeid)

        reportingTimingParams = (0, 10)
        subscription = await client.driver.device_controller.read(
            nodeid, attributes=[clusters.OnOff], reportInterval=reportingTimingParams
        )

        def subscription_event(data):
            _LOGGER.info("Received subscription event %s", data)

        subscription.handler = subscription_event

        # This (now) throws exceptions if it fails
        # await client.driver.device_controller.ResolveNode(nodeid)

        # node = await client.driver.device_controller.Read(
        #     nodeid, attributes=[clusters.OnOff], events="*", returnClusterObject=True
        # )
        # await asyncio.sleep(1)

        # node = await client.driver.device_controller.Read(
        #     nodeid, attributes="*", events="*", returnClusterObject=True
        # )

        # from pprint import pprint

        # pprint(node)

        while True:
            await asyncio.sleep(5)
            await client.driver.device_controller.send_command(
                nodeid=nodeid, endpoint=1, payload=clusters.OnOff.Commands.Toggle()
            )
    except Exception:
        _LOGGER.exception("No more Happiness")


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Coroutine, Generic, TypeVar

from matter_server.vendor import device_types
from matter_server.vendor.chip.clusters import Objects as all_clusters

if TYPE_CHECKING:
    from .node import MatterNode

SubscriberType = Callable[[], None]


_DEVICE_TYPE_T = TypeVar("_DEVICE_TYPE_T", bound=device_types.DeviceType)
_CLUSTER_T = TypeVar("_CLUSTER_T", bound=all_clusters.Cluster)


class MatterDeviceTypeInstance(Generic[_DEVICE_TYPE_T]):
    """Base class for Matter device types on endpoints."""

    def __init__(
        self,
        node: MatterNode,
        device_type: _DEVICE_TYPE_T,
        endpoint_id: int,
        device_revision: int,
    ) -> None:
        self.node = node
        self.device_type = device_type
        self.device_revision = device_revision
        self.endpoint_id = endpoint_id
        self._on_update_listener: SubscriberType | None = None

    @property
    def data(self) -> dict:
        return self.node.raw_data["attributes"][str(self.endpoint_id)]

    def has_cluster(self, cluster: type[all_clusters.Cluster]) -> bool:
        """Check if device has a specific cluster."""
        return cluster in self.device_type.clusters and cluster.__name__ in self.data

    def get_cluster(self, cluster: type[_CLUSTER_T]) -> _CLUSTER_T | None:
        """Get the cluster object."""
        if not self.has_cluster(cluster):
            return None

        return self.data[cluster.__name__]

    async def send_command(
        self,
        payload: all_clusters.ClusterCommand,
        responseType=None,
        timedRequestTimeoutMs: int = None,
    ):
        """Send a command to the device."""
        return await self.node.matter.client.driver.device_controller.send_command(
            nodeid=self.node.node_id,
            endpoint=self.endpoint_id,
            payload=payload,
            responseType=responseType,
            timedRequestTimeoutMs=timedRequestTimeoutMs,
        )

    async def update_attributes(self, attributes: list) -> None:
        """Update attributes."""
        result = await self.node.matter.client.driver.device_controller.read(
            self.node.node_id,
            attributes=[(self.endpoint_id, attribute) for attribute in attributes],
        )

        # {
        #     "attributes": {
        #         "1": {
        #             "LevelControl": {"DataVersion": 2821513409, "CurrentLevel": 1},
        #             "OnOff": {"DataVersion": 781696941, "OnOff": True},
        #         }
        #     },
        #     "events": [],
        #     "_type": "chip.clusters.Attribute.AsyncReadTransaction.ReadResponse",
        # }

        self.node.matter.adapter.logger.debug("Read result: %s", result)

        updated_data = result["attributes"].get(str(self.endpoint_id))

        for cluster_name, info in updated_data.items():
            info.pop("DataVersion")
            # Convert to dictionary where all the keys start with lowercase
            info = {key[0].lower() + key[1:]: value for key, value in info.items()}
            for attribute, value in info.items():
                self.node.matter.adapter.logger.debug(
                    "node %s, endpoint %s: updating %s, set %s=%s",
                    self.node.node_id,
                    self.endpoint_id,
                    cluster_name,
                    attribute,
                    value,
                )
                setattr(self.data[cluster_name], attribute, value)

    async def subscribe_updates(
        self, subscribe_attributes: list, subscriber: SubscriberType
    ) -> Callable[[], Coroutine[None]]:
        """Subscribe to updates."""
        if self._on_update_listener is not None:
            raise RuntimeError("Cannot subscribe twice!")

        self.node.matter.adapter.logger.debug(
            "node %s, endpoint %s: subscribing to %s",
            self.node.node_id,
            self.endpoint_id,
            subscribe_attributes,
        )

        self._on_update_listener = subscriber

        reporting_timing_params = (0, 10)

        unsub_subscription = (
            await self.node.matter.client.driver.read_subscriptions.subscribe_node(
                nodeid=self.node.node_id,
                subscription_callback=self._receive_event,
                attributes=[
                    (self.endpoint_id, attribute) for attribute in subscribe_attributes
                ],
                reportInterval=reporting_timing_params,
                keepSubscriptions=True,
            )
        )

        async def unsubscribe() -> None:
            self.node.matter.adapter.logger.debug(
                "node %s, endpoint %s: unsubscribing from %s",
                self.node.node_id,
                self.endpoint_id,
                subscribe_attributes,
            )
            self._on_update_listener = None
            await unsub_subscription()

        return unsubscribe

    def _receive_event(self, event):
        self.node.matter.adapter.logger.debug(
            "node %s, endpoint %s: received subscription event %s",
            self.node.node_id,
            self.endpoint_id,
            event,
        )

        # 2022-06-03 11:30:52 DEBUG (MainThread) [custom_components.matter_experimental.adapter] node 4335, endpoint 1: received subscription event
        # {'SubscriptionId': 2632459106, 'FabridId': 1, 'NodeId': 4335, 'Endpoint': 1, 'Attribute': <class 'matter_server.vendor.chip.clusters.Objects.OnOff.Attributes.OnOff'>, 'Value': False}

        self.node.matter.adapter.logger.debug(
            "Updating node %s, endpoint %s, %s: %s=%s",
            self.node.node_id,
            self.endpoint_id,
            event["cluster"].__name__,
            event["attribute"],
            event["value"],
        )

        setattr(
            self.data[event["cluster"].__name__], event["attribute"], event["value"]
        )

        if self._on_update_listener:
            self._on_update_listener()

    def __repr__(self):
        return f"<MatterDeviceTypeInstance {self.device_type.__name__} (N:{self.node.node_id}, E:{self.endpoint_id})>"



"""Matter base class."""
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Callable

from matter_server.vendor.chip.clusters.ObjectsVersion import CLUSTER_OBJECT_VERSION

from .client import Client
from .exceptions import BaseMatterServerError, FailedCommand
from .model.node import MatterNode

if TYPE_CHECKING:
    from .adapter import AbstractMatterAdapter


INTERVIEW_RETRY_TIME = 60  # seconds


class Matter:

    next_node_id: int
    _nodes: dict[int, MatterNode | None]
    _listen_task: asyncio.Task | None = None
    _handle_driver_ready_task: asyncio.Task | None = None
    _reinterview_unsub: Callable[[], None] | None = None
    _reinterview_nodes: set[int] | None = None
    _commission_lock: asyncio.Lock | None = None

    def __init__(
        self,
        adapter: AbstractMatterAdapter,
    ):
        self.adapter = adapter
        self.client = Client(adapter.get_server_url(), adapter.get_client_session())
        self.driver_ready = asyncio.Event()

    def get_node(self, node_id: int) -> MatterNode:
        """Get a node."""
        node = self._nodes[node_id]
        if node is None:
            raise ValueError(f"Node {node_id} not interviewed")
        return node

    def get_nodes(self) -> list[MatterNode]:
        """Get nodes."""
        # node is None if not interviewed
        return [node for node in self._nodes.values() if node]

    async def connect(self):
        """Connect to the server."""
        await self.client.connect()

        data = await self.adapter.load_data()

        if (
            data
            and data.get("compressed_fabric_id")
            != self.client.server_info.compressedFabricId
        ):
            self.adapter.logger.warning(
                "Connected to a server with a new fabric ID (current: %s, server: %s). Resetting data",
                data.get("compressed_fabric_id"),
                self.client.server_info.compressedFabricId,
            )
            data = None
            # TODO can we detect all known nodes to the server and interview them?

        if data is None:
            data = {"next_node_id": 4335, "nodes": {}}
        elif data.get("node_interview_version") != CLUSTER_OBJECT_VERSION:
            data["nodes"] = {key: None for key in data["nodes"]}

        self.next_node_id = data["next_node_id"]
        # JSON stores dictionary keys as strings, convert to int
        self._nodes = {
            int(node_id): MatterNode(self, node_info) if node_info else None
            for node_id, node_info in data["nodes"].items()
        }

    async def disconnect(self):
        """Disconnect from the server."""
        self._listen_task.cancel()
        await self._listen_task

    async def commission(self, code: str):
        """Commission a new device."""
        if self._commission_lock is None:
            self._commission_lock = asyncio.Lock()

        async with self._commission_lock:
            node_id = self.next_node_id
            await self.client.driver.device_controller.commission_with_code(
                code, node_id
            )
            self._nodes[node_id] = None
            self.next_node_id += 1
            # Save right away because we don't want to lose node IDs
            await self.adapter.save_data(self._data_to_save())

        await self._interview_node(node_id)

    async def delete_node(self, node_id: int) -> None:
        """Delete a node."""
        # TODO notify anyone using MatterNode
        # TODO can we factory reset node_id so it can be commissioned again?
        self._nodes.pop(node_id)
        self.adapter.delay_save_data(self._data_to_save)

    async def _interview_node(self, node_id: int) -> None:
        """Interview a node."""
        self.adapter.logger.info("Interviewing node %s", node_id)
        try:
            await self.client.driver.device_controller.resolve_node(node_id)
            node_info = await self.client.driver.device_controller.read(
                node_id, attributes="*", events="*", returnClusterObject=True
            )
        except FailedCommand as err:
            self.adapter.logger.error("Failed to interview node: %s", err)
            self._schedule_interview_retry({node_id})
            return

        node_info["node_id"] = node_id

        if self._nodes[node_id]:
            # TODO this updates data but doesn't re-run init work or update anyone
            self._nodes[node_id].update_data(node_info)
        else:
            self._nodes[node_id] = MatterNode(self, node_info)

        self.adapter.delay_save_data(self._data_to_save)

        try:
            await self.adapter.setup_node(self._nodes[node_id])
        except Exception:  # pylint: disable=broad-except
            self.adapter.logger.exception(
                "Unexptected error setting up node %s", node_id
            )

    def _schedule_interview_retry(self, nodes: set[int], timeout=INTERVIEW_RETRY_TIME):
        """Schedule a retry of failed nodes."""
        if self._reinterview_unsub is None:
            loop = asyncio.get_running_loop()
            self._reinterview_unsub = loop.call_later(
                timeout, lambda: loop.create_task(self._interview_nodes_retry())
            ).cancel

        if self._reinterview_nodes is None:
            self._reinterview_nodes = nodes
        else:
            self._reinterview_nodes |= nodes

    async def _interview_nodes_retry(self) -> None:
        """Retry failed nodes."""
        # Keep track of tried to guard re-interviewing the same node
        tried: set[int] = set()
        assert isinstance(self._reinterview_nodes, set)

        while to_handle := self._reinterview_nodes - tried:
            node_id = to_handle.pop()
            tried.add(node_id)
            await self._interview_node(node_id)
            self._reinterview_nodes.discard(node_id)

        self._reinterview_unsub = None

        if self._reinterview_nodes:
            self._schedule_interview_retry(self._reinterview_nodes)
        else:
            self._reinterview_nodes = None

    def listen(self):
        """Start listening to changes."""
        self._listen_task = asyncio.create_task(self._client_listen())
        self._handle_driver_ready_task = asyncio.create_task(
            self._handle_driver_ready()
        )

    async def _client_listen(self) -> None:
        """Listen with the client."""
        should_reload = True
        try:
            await self.client.listen(self.driver_ready)
        except asyncio.CancelledError:
            should_reload = False
        except BaseMatterServerError as err:
            self.adapter.logger.error("Failed to listen: %s", err)
        except Exception as err:  # pylint: disable=broad-except
            # We need to guard against unknown exceptions to not crash this task.
            self.adapter.logger.exception("Unexpected exception: %s", err)

        self._handle_driver_ready_task.cancel()

        await self.adapter.handle_server_disconnected(should_reload)

    async def _handle_driver_ready(self) -> None:
        """Handle driver ready."""
        await self.driver_ready.wait()
        nodes = self.get_nodes()
        tasks = [self.adapter.setup_node(node) for node in nodes]

        if tasks:
            results = await asyncio.gather(*tasks, return_exceptions=True)

            for node, result in zip(nodes, results):
                if isinstance(result, Exception):
                    self.adapter.logger.error(
                        "Unexpected error setting up node %s",
                        node.node_id,
                        exc_info=result,
                    )

        to_interview = {
            node_id for node_id, info in self._nodes.items() if info is None
        }
        if not to_interview:
            return

        self.adapter.logger.info("Nodes that still need interviewing: %s", to_interview)
        self._schedule_interview_retry(to_interview, 0)

    def _data_to_save(self) -> dict:
        return {
            "compressed_fabric_id": self.client.server_info.compressedFabricId,
            "next_node_id": self.next_node_id,
            "node_interview_version": CLUSTER_OBJECT_VERSION,
            "nodes": {
                node_id: node.raw_data if node else None
                for node_id, node in self._nodes.items()
            },
        }



"""Matter adapter."""
from __future__ import annotations

from abc import abstractmethod
import logging
from typing import Callable

import aiohttp

from .model.node import MatterNode


class AbstractMatterAdapter:

    logger: logging.Logger

    @abstractmethod
    async def load_data(self) -> dict | None:
        """Load data."""

    @abstractmethod
    async def save_data(self, data: dict) -> None:
        """Save data."""

    @abstractmethod
    def delay_save_data(self, get_data: Callable[[], dict]) -> None:
        """Save data, but not right now."""

    @abstractmethod
    def get_server_url(self) -> str:
        """Return server URL."""

    @abstractmethod
    def get_client_session(self) -> aiohttp.ClientSession:
        """Return aiohttp client session."""

    @abstractmethod
    async def setup_node(self, node: MatterNode) -> None:
        """Set up a node."""

    @abstractmethod
    async def handle_server_disconnected(self, should_reload: bool) -> None:
        """Server disconnected."""



from typing import Callable


class Subscription:
    subscription_id: int
    handler: Callable[[dict], None]

    def __init__(self, subscription_id):
        self.subscription_id = subscription_id



#
#    Copyright (c) 2020 Project CHIP Authors
#    All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

class Nullable():
    def __repr__(self):
        return 'Null'

    def __eq__(self, other):
        if isinstance(other, Nullable):
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return True


NullValue = Nullable()



from __future__ import annotations

from typing import TYPE_CHECKING

from .device_controller import DeviceController
from .read_subscriptions import ReadSubscriptions

if TYPE_CHECKING:
    from .. import client


class Driver:
    def __init__(self, client: client.Client):
        self.server_info = client.server_info
        self._client = client
        self.device_controller = DeviceController(client)
        self.read_subscriptions = ReadSubscriptions(client)



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

from aiohttp import web

from .websocket import mount_websocket

if TYPE_CHECKING:
    from .matter_stack import MatterStack


class MatterServer:
    """Serve Matter over WS."""

    def __init__(self, stack: MatterStack):
        self.stack = stack
        self.logger = logging.getLogger(__name__)
        self.app = web.Application()
        self.loop = asyncio.get_running_loop()
        mount_websocket(self, "/chip_ws")

    def run(self, host, port):
        """Run the server."""
        web.run_app(self.app, host=host, port=port, loop=self.loop)



import base64
import copy
from dataclasses import fields, is_dataclass
from enum import Enum
import json

import matter_server.common.model

# Compatible both with vendorized and not-vendorized
try:
    from chip.clusters import Objects as Clusters
    from chip.clusters.Types import Nullable
except ImportError:
    from ..vendor.chip.clusters import Objects as Clusters
    from ..vendor.chip.clusters.Types import Nullable

MATTER_SERVER_NAMESPACE = "matter_server.common.model"
CLUSTER_TYPE_NAMESPACE = "chip.clusters.Objects"
CLUSTER_ATTRIBUTE_NAMESPACE = "chip.clusters.Attribute"
CLUSTER_TYPE_VENDORIZED_NAMESPACE = "matter_server.vendor"


def strip_vendorized_namespace(namespace):
    if namespace.startswith(CLUSTER_TYPE_VENDORIZED_NAMESPACE):
        return namespace.removeprefix(f"{CLUSTER_TYPE_VENDORIZED_NAMESPACE}.")
    return namespace


# Copy of dataclasses._asdict_inner
def asdict_typed(obj, dict_factory):
    if is_dataclass(obj) and not isinstance(obj, type):
        cls = type(obj)
        result = []
        for f in fields(obj):
            value = asdict_typed(getattr(obj, f.name), dict_factory)
            result.append((f.name, value))
        # Add type information to prepare for serializing dataclasses
        result.append(
            (
                "_type",
                f"{strip_vendorized_namespace(cls.__module__)}.{cls.__qualname__}",
            )
        )
        return dict_factory(result)
    elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
        return type(obj)(*[asdict_typed(v, dict_factory) for v in obj])
    elif isinstance(obj, (list, tuple)):
        return type(obj)(asdict_typed(v, dict_factory) for v in obj)
    elif isinstance(obj, dict):
        return type(obj)(
            (asdict_typed(k, dict_factory), asdict_typed(v, dict_factory))
            for k, v in obj.items()
        )
    else:
        return copy.deepcopy(obj)


class CHIPJSONEncoder(json.JSONEncoder):
    def _encode_dataclasses_and_dict_keys(self, obj):
        obj = asdict_typed(obj, dict)

        # Matter uses Cluster Attribute classes as dict keys.
        # Arbitrary types as not allowed by JSON encoder
        # Encode as string for now.
        if isinstance(obj, dict):
            return {
                k.__name__
                if isinstance(k, type)
                else k: self._encode_dataclasses_and_dict_keys(v)
                for k, v in obj.items()
            }
        return obj

    def encode(self, obj):
        obj = self._encode_dataclasses_and_dict_keys(obj)
        return json.JSONEncoder.encode(self, obj)

    def default(self, obj):
        if isinstance(obj, Nullable):
            return None
        elif isinstance(obj, Exception):
            # Work around for InteractionModelError/ValueError exceptions currently experiencing in
            # the ThreadNetworkDiagnostics Cluster
            return str(obj)
        elif isinstance(obj, type):
            namespace = strip_vendorized_namespace(obj.__module__)
            # Maybe we should restrict this to CHIP cluster types (see deserialization?)
            return {"_class": f"{namespace}.{obj.__qualname__}"}
        elif isinstance(obj, Enum):
            # Works for chip.clusters.Attributes.EventPriority,
            # might need more sophisticated solution for other Enums
            # Also, deserialization?
            return obj.value
        elif isinstance(obj, bytes):
            return {"_type": "bytes", "value": base64.b64encode(obj).decode("utf-8")}

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


class CHIPJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def _get_class_from_module(self, cluster_cls, type: str):
        for cluster_subtype in type.split("."):
            cluster_cls = getattr(cluster_cls, cluster_subtype)

        return cluster_cls

    def _get_class(self, type: str):
        if type.startswith(CLUSTER_TYPE_NAMESPACE):
            cluster_type = type.removeprefix(f"{CLUSTER_TYPE_NAMESPACE}.")
            return self._get_class_from_module(Clusters, cluster_type)
        elif type.startswith(MATTER_SERVER_NAMESPACE):
            matter_server_type = type.removeprefix(f"{MATTER_SERVER_NAMESPACE}.")
            return self._get_class_from_module(
                matter_server.common.model, matter_server_type
            )
        elif type.startswith(CLUSTER_ATTRIBUTE_NAMESPACE):
            # TODO: How do we deal with that?
            # chip.clusters.Attribute.AsyncReadTransaction.ReadResponse
            return dict
        else:
            raise TypeError(
                f"Type {type} unsupported. Only CHIP cluster and Matter server types supported"
            )

    def object_hook(self, obj: dict):
        if type := obj.get("_type"):
            if type == "bytes":
                return base64.b64decode(obj["value"])
            else:
                cls = self._get_class(type)
                # Delete the `_type` key as it isn't used in the dataclasses
                del obj["_type"]
                return cls(**obj)
        elif cls := obj.get("_class"):
            return self._get_class(cls)

        return obj



from __future__ import annotations

import asyncio
from concurrent import futures
from functools import partial
import logging
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Final, cast

from chip.exceptions import ChipStackError

from ..common.model.message import (
    CommandMessage,
    ErrorResultMessage,
    ServerInformation,
    SubscriptionReportMessage,
    SuccessResultMessage,
)
from ..common.model.version import VersionInfo

if TYPE_CHECKING:
    from chip.clusters import Attribute

    from .server import MatterServer

MAX_PENDING_MSG = 512
CANCELLATION_ERRORS: Final = (asyncio.CancelledError, futures.CancelledError)


class Commands(dict):
    """Small helper to register commands."""

    def register(self, command: str):
        """Decorate a command handler."""

        def decorator(func: Callable[..., Coroutine[Any, Any, Any]]):
            self[command] = func
            return func

        return decorator


class Disconnect(Exception):
    """When the connection needs to be closed."""


class ActiveConnection:
    """Active connection between a client and Matter."""

    commands = Commands()
    fabric_id: int

    def __init__(
        self,
        logger: logging.Logger,
        server: MatterServer,
        send_message: Callable[[Any], None],
    ):
        self.logger = logger
        self.server = server
        self.loop = asyncio.get_running_loop()
        self._send_message = send_message
        self._subscriptions: dict[int, Attribute.SubscriptionTransaction] = {}

    async def async_initialize(self):
        """Initialize connection."""
        self._send_message(
            VersionInfo(
                driver_version=0,
                server_version=0,
                min_schema_version=1,
                max_schema_version=1,
            )
        )

    def async_handle_close(self):
        """Handle connection shutting down."""
        for subscriptionid, subscription in self._subscriptions.items():
            self.logger.info("Shutdown subscription id %d", subscriptionid)
            subscription.Shutdown()

    def async_handle(self, msg: CommandMessage):
        """Handle a command."""
        handler = self.commands.get(msg.command)

        if handler is None:
            self._send_message(ErrorResultMessage(msg.messageId, "INVALID_COMMAND"))
            self.logger.warning("Unknown command: %s", msg.command)
            return

        self.loop.create_task(self._run_handler(handler, msg))

    async def _run_handler(self, handler, msg: CommandMessage):
        try:
            await handler(self, msg)
        except ChipStackError as ex:
            self._send_message(ErrorResultMessage(msg.messageId, str(ex)))
        except Exception:  # pylint: disable=broad-except
            self.logger.exception("Error handling message: %s", msg)
            self._send_message(ErrorResultMessage(msg.messageId, "unknown_error"))

    # pylint: disable=invalid-name

    @commands.register("server.GetInfo")
    async def _handle_server_GetInfo(self, msg: CommandMessage):
        self._send_message(
            SuccessResultMessage(
                msg.messageId,
                ServerInformation(
                    fabricId=self.server.stack.fabric_id,
                    compressedFabricId=self.server.stack.compressed_fabric_id,
                ),
            )
        )

    @commands.register("device_controller.CommissionWithCode")
    async def _handle_device_controller_CommissionWithCode(self, msg: CommandMessage):
        if not self.server.stack.wifi_cred_set:
            self.logger.warning("Received commissioning without Wi-Fi set")

        result = await self.loop.run_in_executor(
            None,
            partial(self.server.stack.device_controller.CommissionWithCode, **msg.args),
        )
        self._send_message(SuccessResultMessage(msg.messageId, {"raw": result}))

    @commands.register("device_controller.Read")
    async def _handle_device_controller_Read(self, msg: CommandMessage):
        if isinstance(msg.args.get("attributes"), list):
            converted_attributes = []
            for attribute in msg.args["attributes"]:
                if isinstance(attribute, list):
                    converted_attributes.append(tuple(attribute))
                else:
                    converted_attributes.append(attribute)

            msg.args["attributes"] = converted_attributes

        result = await self.server.stack.device_controller.Read(**msg.args)

        if msg.args.get("reportInterval") is None:
            self._send_message(SuccessResultMessage(msg.messageId, result))
            return

        self.logger.info("Setting up Subscription for %s", msg.args["attributes"])

        subscription = cast("Attribute.SubscriptionTransaction", result)

        # pylint: disable=protected-access
        subscription_id = subscription._subscriptionId

        # Subscription, we need to keep track on it server side
        def subscription_callback(
            path: Attribute.TypedAttributePath,
            subscription: Attribute.SubscriptionTransaction,
        ):
            data = subscription.GetAttribute(path)
            value = {
                "subscriptionId": subscription_id,
                "fabridId": self.server.stack.fabric_id,
                "nodeId": msg.args["nodeid"],
                "endpoint": path.Path.EndpointId,
                "cluster": path.ClusterType,
                "attribute": path.AttributeName,
                "value": data,
            }
            self.logger.info("subscription_callback %s", value)

            # This callback is running in the CHIP stack thread
            self.loop.call_soon_threadsafe(
                self._send_message, SubscriptionReportMessage(subscription_id, value)
            )

        subscription.SetAttributeUpdateCallback(subscription_callback)
        self.logger.info(f"SubscriptionId {subscription_id}")
        self._subscriptions[subscription_id] = subscription
        self._send_message(
            SuccessResultMessage(msg.messageId, {"subscription_id": subscription_id})
        )

    @commands.register("device_controller.ResolveNode")
    async def _handle_device_controller_ResolveNode(self, msg: CommandMessage):
        result = await self.loop.run_in_executor(
            None, partial(self.server.stack.device_controller.ResolveNode, **msg.args)
        )
        self._send_message(SuccessResultMessage(msg.messageId, {"raw": result}))

    @commands.register("device_controller.SendCommand")
    async def _handle_device_controller_SendCommand(self, msg: CommandMessage):
        result = await self.server.stack.device_controller.SendCommand(**msg.args)
        self._send_message(SuccessResultMessage(msg.messageId, {"raw": result}))

    @commands.register("device_controller.SetThreadOperationalDataset")
    async def _handle_device_controller_SetThreadOperationalDataset(
        self, msg: CommandMessage
    ):
        result = await self.loop.run_in_executor(
            None,
            partial(
                self.server.stack.device_controller.SetThreadOperationalDataset,
                **msg.args,
            ),
        )
        self._send_message(SuccessResultMessage(msg.messageId, {"raw": result}))

    @commands.register("device_controller.OpenCommissioningWindow")
    async def _handle_device_controller_OpenCommissioningWindow(
        self, msg: CommandMessage
    ):
        result = await self.loop.run_in_executor(
            None,
            partial(
                self.server.stack.device_controller.OpenCommissioningWindow,
                **msg.args,
            ),
        )
        self._send_message(SuccessResultMessage(msg.messageId, {"raw": result}))

    @commands.register("device_controller.SetWiFiCredentials")
    async def _handle_device_controller_SetWiFiCredentials(self, msg: CommandMessage):
        result = await self.loop.run_in_executor(
            None,
            partial(self.server.stack.device_controller.SetWiFiCredentials, **msg.args),
        )

        if result == 0:
            self.server.stack.wifi_cred_set = True

        self._send_message(SuccessResultMessage(msg.messageId, {"raw": result}))

    @commands.register("device_controller.Unsubscribe")
    async def _handle_device_controller_Unsubscribe(self, msg: CommandMessage):
        """Unsubscribe."""
        subscription = self._subscriptions.pop(msg.args["subscription_id"], None)
        if subscription is None:
            self._send_message(ErrorResultMessage(msg.messageId, "not_found"))
            return

        await self.loop.run_in_executor(None, subscription.Shutdown)
        self._send_message(SuccessResultMessage(msg.messageId, None))



import logging

from chip.ChipDeviceCtrl import ChipDeviceController
from chip.ChipStack import ChipStack
from chip.FabricAdmin import FabricAdmin
import chip.logging
import chip.native

_LOGGER = logging.getLogger(__name__)


class MatterStack:

    # To track if wifi credentials set this session.
    wifi_cred_set = False
    device_controller: ChipDeviceController

    def __init__(self, storage_path):
        _LOGGER.info("Setup CHIP Controller Server")
        chip.native.GetLibraryHandle()
        chip.logging.RedirectToPythonLogging()
        self.stack = ChipStack(persistentStoragePath=storage_path)
        self.fabric_admin = FabricAdmin()
        self.device_controller = self.fabric_admin.NewController()
        _LOGGER.info("CHIP Controller Stack initialized")
        self.fabric_id = self.device_controller.GetFabricId()
        self.compressed_fabric_id = self.device_controller.GetCompressedFabricId()

    def shutdown(self):
        _LOGGER.info("Shutdown CHIP Controller Server")
        self.device_controller.Shutdown()
        self.stack.Shutdown()



"""Represents the version from the server."""

from dataclasses import dataclass


@dataclass
class VersionInfo:
    """Version info of the server."""

    driver_version: str
    server_version: str
    min_schema_version: int
    max_schema_version: int



from __future__ import annotations

import logging
import typing
from typing import TYPE_CHECKING

from matter_server.client.model.subscription import Subscription
from matter_server.vendor.chip.clusters import Objects as ClusterObjects

_LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .. import client


ReadAttributesType = typing.List[
    typing.Union[
        None,  # Empty tuple, all wildcard
        typing.Tuple[int],  # Endpoint
        # Wildcard endpoint, Cluster id present
        typing.Tuple[typing.Type[ClusterObjects.Cluster]],
        # Wildcard endpoint, Cluster + Attribute present
        typing.Tuple[typing.Type[ClusterObjects.ClusterAttributeDescriptor]],
        # Wildcard attribute id
        typing.Tuple[int, typing.Type[ClusterObjects.Cluster]],
        # Concrete path
        typing.Tuple[int, typing.Type[ClusterObjects.ClusterAttributeDescriptor]],
    ]
]
ReadDataVersionFiltersType = typing.List[
    typing.Tuple[int, typing.Type[ClusterObjects.Cluster], int]
]
ReadEventsType = typing.List[
    typing.Union[
        None,  # Empty tuple, all wildcard
        typing.Tuple[str, int],  # all wildcard with urgency set
        typing.Tuple[int, int],  # Endpoint,
        # Wildcard endpoint, Cluster id present
        typing.Tuple[typing.Type[ClusterObjects.Cluster], int],
        # Wildcard endpoint, Cluster + Event present
        typing.Tuple[typing.Type[ClusterObjects.ClusterEvent], int],
        # Wildcard event id
        typing.Tuple[int, typing.Type[ClusterObjects.Cluster], int],
        # Concrete path
        typing.Tuple[int, typing.Type[ClusterObjects.ClusterEvent], int],
    ]
]


class DeviceController:
    def __init__(self, client: client.Client):
        self.client = client
        self.subscriptions: dict[int, Subscription] = {}

    async def commission_with_code(self, setupPayload: str, nodeid: int):
        return await self._async_send_command(
            "CommissionWithCode",
            {
                "setupPayload": setupPayload,
                "nodeid": nodeid,
            },
        )

    async def set_wifi_credentials(self, ssid: str, credentials: str):
        return await self._async_send_command(
            "SetWiFiCredentials",
            {
                "ssid": ssid,
                "credentials": credentials,
            },
        )

    async def set_thread_operational_dataset(self, dataset: bytes):
        return await self._async_send_command(
            "SetThreadOperationalDataset",
            {
                "threadOperationalDataset": dataset,
            },
        )

    async def resolve_node(self, nodeid: int):
        return await self._async_send_command(
            "ResolveNode",
            {
                "nodeid": nodeid,
            },
        )

    async def send_command(
        self,
        nodeid: int,
        endpoint: int,
        payload: ClusterObjects.ClusterCommand,
        responseType=None,
        timedRequestTimeoutMs: int = None,
    ):
        """
        Send a cluster-object encapsulated command to a node and get returned a future that can be awaited upon to receive the response.
        If a valid responseType is passed in, that will be used to deserialize the object. If not, the type will be automatically deduced
        from the metadata received over the wire.

        timedWriteTimeoutMs: Timeout for a timed invoke request. Omit or set to 'None' to indicate a non-timed request.
        """
        # TODO add these only if defined.
        # "responseType": responseType,
        # "timedRequestTimeoutMs": timedRequestTimeoutMs,
        return await self._async_send_command(
            "SendCommand",
            {
                "nodeid": nodeid,
                "endpoint": endpoint,
                "payload": payload,
            },
        )

    async def read(
        self,
        nodeid: int,
        attributes: ReadAttributesType = None,
        dataVersionFilters: ReadDataVersionFiltersType = None,
        events: ReadEventsType = None,
        returnClusterObject: bool = False,
        reportInterval: typing.Tuple[int, int] = None,
        fabricFiltered: bool = True,
        keepSubscriptions: bool = False,
    ):
        """
        Read a list of attributes and/or events from a target node

        nodeId: Target's Node ID
        attributes: A list of tuples of varying types depending on the type of read being requested:
            (endpoint, Clusters.ClusterA.AttributeA):   Endpoint = specific,    Cluster = specific,   Attribute = specific
            (endpoint, Clusters.ClusterA):              Endpoint = specific,    Cluster = specific,   Attribute = *
            (Clusters.ClusterA.AttributeA):             Endpoint = *,           Cluster = specific,   Attribute = specific
            endpoint:                                   Endpoint = specific,    Cluster = *,          Attribute = *
            Clusters.ClusterA:                          Endpoint = *,           Cluster = specific,   Attribute = *
            '*' or ():                                  Endpoint = *,           Cluster = *,          Attribute = *

            The cluster and attributes specified above are to be selected from the generated cluster objects.

            e.g.
                ReadAttribute(1, [ 1 ] ) -- case 4 above.
                ReadAttribute(1, [ Clusters.Basic ] ) -- case 5 above.
                ReadAttribute(1, [ (1, Clusters.Basic.Attributes.Location ] ) -- case 1 above.

        dataVersionFilters: A list of tuples of (endpoint, cluster, data version).

        events: A list of tuples of varying types depending on the type of read being requested:
            (endpoint, Clusters.ClusterA.EventA, urgent):       Endpoint = specific,    Cluster = specific,   Event = specific, Urgent = True/False
            (endpoint, Clusters.ClusterA, urgent):              Endpoint = specific,    Cluster = specific,   Event = *, Urgent = True/False
            (Clusters.ClusterA.EventA, urgent):                 Endpoint = *,           Cluster = specific,   Event = specific, Urgent = True/False
            endpoint:                                   Endpoint = specific,    Cluster = *,          Event = *, Urgent = True/False
            Clusters.ClusterA:                          Endpoint = *,           Cluster = specific,   Event = *, Urgent = True/False
            '*' or ():                                  Endpoint = *,           Cluster = *,          Event = *, Urgent = True/False

        returnClusterObject: This returns the data as consolidated cluster objects, with all attributes for a cluster inside
                             a single cluster-wide cluster object.

        reportInterval: A tuple of two int-s for (MinIntervalFloor, MaxIntervalCeiling). Used by establishing subscriptions.
            When not provided, a read request will be sent.
        """
        return await self._async_send_command(
            "Read",
            {
                "nodeid": nodeid,
                "attributes": attributes,
                "dataVersionFilters": dataVersionFilters,
                "events": events,
                "returnClusterObject": returnClusterObject,
                "reportInterval": reportInterval,
                "fabricFiltered": fabricFiltered,
                "keepSubscriptions": keepSubscriptions,
            },
        )

    async def open_commissioning_window(self, nodeid):
        """Open commissioning window."""
        # TODO: We need to generate a discriminator and tell the user
        # We might also opt to use Basic Commissioning Mode, but
        # it seems to be an optional feature :o)
        # option:
        #  1 to use Enhanced Commissioning Method.
        #  0 to use Basic Commissioning Method.
        return await self._async_send_command(
            "OpenCommissioningWindow",
            {
                "nodeid": nodeid,
                "timeout": 300,
                "iteration": 1000,
                "discriminator": 3840,
                "option": 0,
            },
        )

    async def _async_send_command(self, command, args):
        """Send driver controller command."""
        return await self.client.async_send_command(
            f"device_controller.{command}", args
        )



from dataclasses import dataclass
from typing import Any


@dataclass
class Message:
    pass


@dataclass
class ResultMessage:
    messageId: str
    pass


@dataclass
class SuccessResultMessage(ResultMessage):
    result: Any


@dataclass
class ErrorResultMessage(ResultMessage):
    errorCode: Any


@dataclass
class SubscriptionReportMessage(Message):
    subscriptionId: str
    payload: Any


@dataclass
class CommandMessage(Message):
    messageId: str
    command: str
    args: dict[str, Any]


@dataclass
class ServerInformation:
    fabricId: int
    compressedFabricId: int
